import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;

/**
 * A colourful letter on the screen. It adjusts it's brightness based
 * on an initial angle which is applied to a sin wave to follow brightness
 * adjustments.
 * 
 * Also moves towards the mouse if it is nearby.
 * 
 * @author Joseph Lenton
 */
public class Letter  extends Actor
{
    public static final int WIDTH = 20;
    public static final int HEIGHT = 28;
    
    private static final double MAX_MOUSE_DIST = Math.sqrt( Mouse.WIDTH*Mouse.WIDTH + Mouse.HEIGHT*Mouse.HEIGHT );
    
    private static final double ANGLE_INC = Math.toRadians(5);
    
    private static final int MIN_CHANGE_TIME = 16; // approx 1 frame
    private static final int MAX_CHANGE_TIME = 1000; // approx 60 frames
    private static final int DIFF_CHANGE_TIME = MAX_CHANGE_TIME-MIN_CHANGE_TIME;
    
    private static GreenfootImage[] LETTERS = new GreenfootImage[] {
            new GreenfootImage("images/font/a.png"),
            new GreenfootImage("images/font/b.png"),
            new GreenfootImage("images/font/c.png"),
            new GreenfootImage("images/font/d.png"),
            new GreenfootImage("images/font/e.png"),
            new GreenfootImage("images/font/f.png"),
            new GreenfootImage("images/font/g.png"),
            new GreenfootImage("images/font/h.png"),
            new GreenfootImage("images/font/i.png"),
            new GreenfootImage("images/font/j.png"),
            new GreenfootImage("images/font/k.png"),
            new GreenfootImage("images/font/l.png"),
            new GreenfootImage("images/font/m.png"),
            new GreenfootImage("images/font/n.png"),
            new GreenfootImage("images/font/o.png"),
            new GreenfootImage("images/font/p.png"),
            new GreenfootImage("images/font/q.png"),
            new GreenfootImage("images/font/r.png"),
            new GreenfootImage("images/font/s.png"),
            new GreenfootImage("images/font/t.png"),
            new GreenfootImage("images/font/u.png"),
            new GreenfootImage("images/font/v.png"),
            new GreenfootImage("images/font/w.png"),
            new GreenfootImage("images/font/x.png"),
            new GreenfootImage("images/font/y.png"),
            new GreenfootImage("images/font/z.png"),
            
            new GreenfootImage("images/font/0.png"),
            new GreenfootImage("images/font/1.png"),
            new GreenfootImage("images/font/2.png"),
            new GreenfootImage("images/font/3.png"),
            new GreenfootImage("images/font/4.png"),
            new GreenfootImage("images/font/5.png"),
            new GreenfootImage("images/font/6.png"),
            new GreenfootImage("images/font/7.png"),
            new GreenfootImage("images/font/8.png"),
            new GreenfootImage("images/font/9.png"),
            
            new GreenfootImage("images/font/dot.png"),
            new GreenfootImage("images/font/comma.png"),
            new GreenfootImage("images/font/colon.png"),
            new GreenfootImage("images/font/@.png"),
            new GreenfootImage("images/font/slash.png"),
            new GreenfootImage("images/font/single_quote.png")
    };
    
    private double brightnessAngle;
    private GreenfootImage image;
    private int red;
    private int green;
    private int blue;
    private long letterChange;
    
    private int x;
    private int y;
    
    /**
     * The startAngle is the angle used for initiating the brightness.
     * red, green and blue define the color for this letter.
     * 
     * @param startAngle The initial angle used for calculating this Letters brightness.
     * @param red The red component for this Letter's color.
     * @param green The green component for this Letter's color.
     * @param blue The blue component for this Letter's color.
     */
    public Letter(double startAngle, int red, int green, int blue)
    {
        this.brightnessAngle = startAngle;
        
        this.red = red;
        this.green = green;
        this.blue = blue;
        
        changeRandomImage();
        setImage( new GreenfootImage(WIDTH, HEIGHT) );
        redrawImage();
    }
    
    /**
     * Records it's starting location for future use.
     */
    @Override
    public void addedToWorld(World world)
    {
        super.addedToWorld( world );
        this.x = getX();
        this.y = getY();
    }
    
    /**
     * Cleares the image with black, then fills in this letters color
     * and then draws it's image.
     */
    private final void redrawImage()
    {
        final GreenfootImage img = getImage();
        final double colorMult = Math.sin(brightnessAngle);
        
        img.setColor( Scenario.BACKGROUND_COLOR );
        img.fill();
        final int drawX = (img.getWidth()-image.getWidth())/2;
        final int drawY = (img.getHeight()-image.getHeight())/2;
        
        img.setColor( new Color(
                colorMinMax(red, colorMult),
                colorMinMax(green, colorMult),
                colorMinMax(blue, colorMult)
        ) );
        img.fillRect( drawX, drawY, image.getWidth(), image.getHeight() );
        img.drawImage( image, drawX, drawY );
        
        setImage( img );
    }
    
    /**
     * Sets a random letter image as the image for this Letter.
     * It also updates the letter image timer delay.
     */
    private final void changeRandomImage()
    {
        image = LETTERS[ Greenfoot.getRandomNumber(LETTERS.length) ];
        letterChange = System.currentTimeMillis() + Greenfoot.getRandomNumber(DIFF_CHANGE_TIME) + MIN_CHANGE_TIME;
    }
    
    /**
     * Updates the letters brightness and redraws it's image.
     */
    public void act() 
    {
        long now = System.currentTimeMillis();
        
        if ( now > letterChange ) {
            changeRandomImage();
        }
        
        brightnessAngle += ANGLE_INC;
        
        actApproachMouse();
        redrawImage();
    }
    
    /**
     * Moves this Letter towards the mouse, but only if it is nearby.
     */
    protected void actApproachMouse()
    {
        final Mouse mouse = ((Scenario) getWorld()).getMouse();
        
        if ( mouse != null && mouse.isOnScreen() ) {
            final int xDiff = mouse.getX()-x;
            final int yDiff = mouse.getY()-y;
            final double dist = Math.sqrt( xDiff*xDiff + yDiff*yDiff );
            
            // manual ellipse intersection check
            if ( dist < MAX_MOUSE_DIST ) {
                // some maths to create a multiplier based on distance to mouse
                final double distMult = dist / MAX_MOUSE_DIST;
                final double invertDistMult = 1 - Math.pow(distMult, 0.5);
                
                setLocation( (int)(x+xDiff*invertDistMult), (int)(y+yDiff*invertDistMult) );
            } else {
                setLocation( x, y );
            }
        } else {
            setLocation( x, y );
        }
    }
    
    /**
     * Multiplies the given mult by 128 (half a maximum color value)
     * and adds this to col. The value is capped between 0 (minimum
     * color) and 255 (maximum color), then returned.
     * 
     * @param col The color to alter, cap and then return.
     * @param mult A multiplier against color, expected to be from -1 to 1.
     * @return col adjusted by the given mult, capped to range of 0 to 255 and then returned.
     */
    private int colorMinMax(int col, double mult)
    {
        col = col + (int) (mult*128);
        return Math.max( Math.min(col, 255), 0 );
    }
}
